from future import standard_library
standard_library.install_aliases()
import os  # @IgnorePep8
import os.path as op  # @IgnorePep8
from arcana.data.file_format import text_format  # @IgnorePep8
from arcana.study import Study, StudyMetaClass  # @IgnorePep8
from arcana.data import (  # @IgnorePep8
    Fileset, InputFilesetSpec, FilesetSpec, Field)  # @IgnorePep8
from arcana.utils.testing import BaseMultiSubjectTestCase  # @IgnorePep8
from arcana.repository import Tree  # @IgnorePep8
from future.utils import with_metaclass  # @IgnorePep8
from arcana.utils.testing import BaseTestCase  # @IgnorePep8
from arcana.data.file_format import FileFormat  # @IgnorePep8


# A dummy format that contains a header
with_header_format = FileFormat(name='with_header', extension='.whf',
                                aux_files={'header': '.hdr'})


class DummyStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('source1', text_format),
        InputFilesetSpec('source2', text_format),
        InputFilesetSpec('source3', text_format),
        InputFilesetSpec('source4', text_format,
                            optional=True),
        FilesetSpec('sink1', text_format, 'dummy_pipeline'),
        FilesetSpec('sink3', text_format, 'dummy_pipeline'),
        FilesetSpec('sink4', text_format, 'dummy_pipeline'),
        FilesetSpec('subject_sink', text_format, 'dummy_pipeline',
                    frequency='per_subject'),
        FilesetSpec('visit_sink', text_format, 'dummy_pipeline',
                    frequency='per_visit'),
        FilesetSpec('project_sink', text_format, 'dummy_pipeline',
                    frequency='per_study'),
        FilesetSpec('resink1', text_format, 'dummy_pipeline'),
        FilesetSpec('resink2', text_format, 'dummy_pipeline'),
        FilesetSpec('resink3', text_format, 'dummy_pipeline')]

    def dummy_pipeline(self):
        pass


class TestBasicRepo(BaseTestCase):

    STUDY_NAME = 'local_repo'
    INPUT_FILESETS = {'source1': '1',
                      'source2': '2',
                      'source3': '3',
                      'source4': '4'}

    def test_get_fileset(self):
        pass


class TestDirectoryProjectInfo(BaseMultiSubjectTestCase):
    """
    This unittest tests out that extracting the existing scans and
    fields in a project returned in a Tree object.
    """

    DATASET_CONTENTS = {'ones': 1, 'tens': 10, 'hundreds': 100,
                        'thousands': 1000, 'with_header': {'.': 'main',
                                                           'header': 'header'}}

    def get_tree(self, repo, sync_with_repo=False):  # @ReservedAssignment @IgnorePep8
        filesets = [
            # Subject1
            Fileset('ones', text_format,
                    frequency='per_subject',
                    subject_id='subject1',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    frequency='per_subject',
                    subject_id='subject1',
                    resource_name='text',
                    repository=repo),
            # subject1/visit1
            Fileset('hundreds', text_format,
                    subject_id='subject1', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            Fileset('ones', text_format,
                    subject_id='subject1', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    subject_id='subject1', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            Fileset('with_header', text_format,
                    frequency='per_session',
                    subject_id='subject1', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            # subject1/visit2
            Fileset('ones', text_format,
                    subject_id='subject1', visit_id='visit2',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    subject_id='subject1', visit_id='visit2',
                    resource_name='text',
                    repository=repo),
            # Subject 2
            Fileset('ones', text_format,
                    frequency='per_subject',
                    subject_id='subject2',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    frequency='per_subject',
                    subject_id='subject2',
                    resource_name='text',
                    repository=repo),
            # subject2/visit1
            Fileset('ones', text_format,
                    subject_id='subject2', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    subject_id='subject2', visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            # subject2/visit2
            Fileset('ones', text_format,
                    subject_id='subject2', visit_id='visit2',
                    resource_name='text',
                    repository=repo),
            Fileset('tens', text_format,
                    subject_id='subject2', visit_id='visit2',
                    resource_name='text',
                    repository=repo),
            # Visit 1
            Fileset('ones', text_format,
                    frequency='per_visit',
                    visit_id='visit1',
                    resource_name='text',
                    repository=repo),
            # Study
            Fileset('ones', text_format,
                    frequency='per_study',
                    resource_name='text',
                    repository=repo)]
        fields = [
            # Subject 2
            Field('e', value=3.33333,
                  frequency='per_subject',
                  subject_id='subject2',
                  repository=repo),
            # subject2/visit2
            Field('a', value=22,
                  subject_id='subject2', visit_id='visit2',
                  repository=repo),
            Field('b', value=220,
                  subject_id='subject2', visit_id='visit2',
                  repository=repo),
            Field('c', value='buggy',
                  subject_id='subject2', visit_id='visit2',
                  repository=repo),
            # Subject1
            Field('e', value=4.44444,
                  frequency='per_subject',
                  subject_id='subject1',
                  repository=repo),
            # subject1/visit1
            Field('a', value=1,
                  subject_id='subject1', visit_id='visit1',
                  repository=repo),
            Field('b', value=10,
                  subject_id='subject1', visit_id='visit1',
                  repository=repo),
            Field('d', value=42.42,
                  subject_id='subject1', visit_id='visit1',
                  repository=repo),
            # subject1/visit2
            Field('a', value=2,
                  subject_id='subject1', visit_id='visit2',
                  repository=repo),
            Field('c', value='van',
                  subject_id='subject1', visit_id='visit2',
                  repository=repo),
            # Visit 1
            Field('f', value='dog',
                  frequency='per_visit',
                  visit_id='visit1',
                  repository=repo),
            # Visit 2
            Field('f', value='cat',
                  frequency='per_visit',
                  visit_id='visit2',
                  repository=repo),
            # Study
            Field('g', value=100,
                  frequency='per_study',
                  repository=repo)]
        # Set URI and IDs if necessary for repository type
        if sync_with_repo:
            for fileset in filesets:
                fileset.get()
            for field in fields:
                field.get()
        tree = Tree.construct(self.repository, filesets, fields)
        return tree

    @property
    def input_tree(self):
        return self.get_tree(self.local_repository)

    def test_project_info(self):
        # Add hidden file (i.e. starting with '.') to local repository at
        # project and subject levels to test ignore functionality
        a_subj_dir = os.listdir(self.project_dir)[0]
        open(op.join(op.join(self.project_dir, '.DS_Store')),
             'w').close()
        open(op.join(self.project_dir, a_subj_dir, '.DS_Store'), 'w').close()
        tree = self.repository.tree()
        for node in tree.nodes():
            for fileset in node.filesets:
                fileset.format = text_format
                fileset._resource_name = 'text'
        self.assertEqual(
            tree, self.local_tree,
            "Generated project doesn't match reference:{}"
            .format(tree.find_mismatch(self.local_tree)))



import os
from unittest import TestCase
from nipype.interfaces.utility import Merge, Split
from arcana.data import (
    InputFilesetSpec, FilesetSpec, InputFilesets, FieldSpec)
from arcana.study.base import Study, StudyMetaClass
from arcana.exceptions import (
    ArcanaModulesNotInstalledException, ArcanaError, ArcanaModulesError)
import unittest
from arcana.utils.testing import BaseTestCase, TestMath
from arcana.data.file_format import text_format
from arcana.environment import ModulesEnv
from arcana.processor import SingleProc
from future.utils import with_metaclass
from arcana.environment import BaseRequirement


class DummyRequirement(BaseRequirement):

    def detect_version_str(self):
        try:
            return os.environ[self.name.upper() + '_VERSION']
        except KeyError:
            loaded_modules = ModulesEnv.loaded()
            raise ArcanaError(
                "Did not find {} in environment variables, found '{}'. "
                "The loaded modules are {}"
                .format(self.name.upper() + '_VERSION',
                        "', '".join(os.environ.keys()),
                        ', '.join(loaded_modules)))


first_req = DummyRequirement('firsttestmodule')
second_req = DummyRequirement('secondtestmodule')

try:
    ModulesEnv._run_module_cmd('avail')
except ArcanaModulesNotInstalledException:
    MODULES_NOT_INSTALLED = True
else:
    MODULES_NOT_INSTALLED = False

SKIP_ARGS = (MODULES_NOT_INSTALLED, "Environment modules are not installed")


class TestMathWithReq(TestMath):

    def _run_interface(self, runtime):
        loaded_modules = ModulesEnv.loaded()
        if first_req.name not in loaded_modules:
            raise ArcanaError(
                "First Test module was not loaded in Node")
        if second_req.name not in loaded_modules:
            raise ArcanaError(
                "Second Test module was not loaded in Node")
        return runtime


class RequirementsStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('ones', text_format),
        FilesetSpec('twos', text_format, 'pipeline1'),
        FieldSpec('threes', float, 'pipeline2'),
        FieldSpec('fours', float, 'pipeline2')]

    def pipeline1(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline1',
            desc=("A pipeline that tests loading of requirements"),
            name_maps=name_maps)
        # Convert from DICOM to NIfTI.gz format on input
        maths = pipeline.add(
            "maths", TestMathWithReq(),
            requirements=[first_req.v('0.15.9'), second_req.v('1.0.2')])
        maths.inputs.op = 'add'
        maths.inputs.as_file = True
        maths.inputs.y = 1
        pipeline.connect_input('ones', maths, 'x', text_format)
        pipeline.connect_output('twos', maths, 'z', text_format)
        return pipeline

    def pipeline2(self, **name_maps):
        pipeline = self.new_pipeline(
            name='pipeline2',
            desc=("A pipeline that tests loading of requirements in "
                  "map nodes"),
            name_maps=name_maps)
        # Convert from DICOM to NIfTI.gz format on input
        merge = pipeline.add("merge", Merge(2))
        maths = pipeline.add(
            "maths", TestMathWithReq(), iterfield='x', requirements=[
                first_req.v('0.15.9'), second_req.v('1.0.2')])
        split = pipeline.add('split', Split())
        split.inputs.splits = [1, 1]
        split.inputs.squeeze = True
        maths.inputs.op = 'add'
        maths.inputs.y = 2
        pipeline.connect_input('ones', merge, 'in1', text_format)
        pipeline.connect_input('twos', merge, 'in2', text_format)
        pipeline.connect(merge, 'out', maths, 'x')
        pipeline.connect(maths, 'z', split, 'inlist')
        pipeline.connect_output('threes', split, 'out1', text_format)
        pipeline.connect_output('fours', split, 'out2', text_format)
        return pipeline


class TestModuleLoad(BaseTestCase):

    INPUT_FILESETS = {'ones': '1'}

    @property
    def processor(self):
        return SingleProc(
            self.work_dir)

    @unittest.skipIf(*SKIP_ARGS)
    def test_module_load(self):
        study = self.create_study(
            RequirementsStudy, 'requirements',
            {'ones': 'ones'},
            environment=ModulesEnv())
        self.assertContentsEqual(study.data('twos'), 2.0)
        self.assertEqual(ModulesEnv.loaded(), {})

    @unittest.skipIf(*SKIP_ARGS)
    def test_module_load_in_map(self):
        study = self.create_study(
            RequirementsStudy, 'requirements',
            [InputFilesets('ones', 'ones', text_format)],
            environment=ModulesEnv())
        threes = study.data('threes')
        fours = study.data('fours')
        self.assertEqual(next(iter(threes)).value, 3)
        self.assertEqual(next(iter(fours)).value, 4)
        self.assertEqual(ModulesEnv.loaded(), {})


class TestModulesRun(TestCase):

    @unittest.skipIf(*SKIP_ARGS)
    def test_run_cmd(self):
        ModulesEnv._run_module_cmd('avail')
        self.assertRaises(
            ArcanaModulesError,
            ModulesEnv._run_module_cmd,
            'badcmd')
        self.assertRaises(
            ArcanaModulesError,
            ModulesEnv._run_module_cmd,
            'load', 'somereallyunlikelymodulename')



import tempfile  # @IgnorePep8
import shutil  # @IgnorePep8
import os
import os.path as op # @IgnorePep8
import unittest  # @IgnorePep8
from unittest import TestCase  # @IgnorePep8
from nipype.interfaces.utility import IdentityInterface  # @IgnorePep8
from arcana.utils.testing import BaseTestCase, BaseMultiSubjectTestCase  # @IgnorePep8
from arcana.study.base import Study, StudyMetaClass  # @IgnorePep8
from arcana.study.parameter import SwitchSpec  # @IgnorePep8
from arcana.data import InputFilesetSpec, FilesetSpec, FieldSpec, InputFilesets  # @IgnorePep8
from arcana.data.file_format import text_format, FileFormat  # @IgnorePep8
from arcana.exceptions import ArcanaDesignError, ArcanaError # @IgnorePep8
from future.utils import PY2  # @IgnorePep8
from future.utils import with_metaclass  # @IgnorePep8
import pydicom  # @IgnorePep8
if PY2:
    import pickle as pkl  # @UnusedImport
else:
    import pickle as pkl  # @Reimport

# For testing DICOM tag matching


class DicomFormat(FileFormat):

    SERIES_NUMBER_TAG = ('0020', '0011')

    def extract_id(self, fileset):
        return int(fileset.dicom_values([self.SERIES_NUMBER_TAG])[0])

    def dicom_values(self, fileset, tags):
        """
        Returns a dictionary with the DICOM header fields corresponding
        to the given tag names

        Parameters
        ----------
        tags : List[Tuple[str, str]]
            List of DICOM tag values as 2-tuple of strings, e.g.
            [('0080', '0020')]
        repository_login : <repository-login-object>
            A login object for the repository to avoid having to relogin
            for every dicom_header call.

        Returns
        -------
        dct : Dict[Tuple[str, str], str|int|float]
        """
        try:
            if (fileset._path is None and fileset._repository is not None and
                    hasattr(fileset.repository, 'dicom_header')):
                hdr = fileset.repository.dicom_header(fileset)
                if not hdr:
                    raise ArcanaError(
                        "No DICOM tags retrieved from {} by {}".format(
                            fileset.repository, fileset))
                values = [hdr[t] for t in tags]
            else:
                # Get the DICOM object for the first file in the fileset
                dcm_files = [f for f in os.listdir(fileset.path)
                             if f.endswith('.dcm')]
                dcm = pydicom.dcmread(op.join(fileset.path, dcm_files[0]))
                values = [dcm[t].value for t in tags]
        except KeyError as e:
            fileset.repository.dicom_header(fileset)
            raise ArcanaError("{} does not have dicom tag {}".format(
                              self, str(e)))
        return values


dicom_format = DicomFormat(name='dicom', extension=None,
                           resource_names={'xnat': ['DICOM']},
                           directory=True, within_dir_exts=['.dcm'])


class TestFilesetSpecPickle(TestCase):

    filesets = []
    fields = []

    def setUp(self):
        self.tmp_dir = tempfile.mkdtemp()
        self.pkl_dir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmp_dir)
        shutil.rmtree(self.pkl_dir)

    def test_fileset_and_field(self):
        objs = [FilesetSpec('a', text_format,
                            'dummy_pipeline1'),
                FieldSpec('b', int, 'dummy_pipeline2')]
        for i, obj in enumerate(objs):
            fname = op.join(self.pkl_dir, '{}.pkl'.format(i))
            with open(fname, 'wb') as f:
                pkl.dump(obj, f)
            with open(fname, 'rb') as f:
                re_obj = pkl.load(f)
            self.assertEqual(obj, re_obj)


class TestMatchStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('gre_phase', dicom_format),
        InputFilesetSpec('gre_mag', dicom_format)]

    def dummy_pipeline1(self):
        pass

    def dummy_pipeline2(self):
        pass


class TestFilesetSelecting(BaseMultiSubjectTestCase):

    @unittest.skip("Test not implemented")
    def test_match_pattern(self):
        pass


class TestDicomTagMatch(BaseTestCase):

    IMAGE_TYPE_TAG = ('0008', '0008')
    GRE_PATTERN = 'gre_field_mapping_3mm.*'
    PHASE_IMAGE_TYPE = ['ORIGINAL', 'PRIMARY', 'P', 'ND']
    MAG_IMAGE_TYPE = ['ORIGINAL', 'PRIMARY', 'M', 'ND', 'NORM']
    DICOM_MATCH = [
        InputFilesets('gre_phase', GRE_PATTERN, dicom_format,
                     dicom_tags={IMAGE_TYPE_TAG: PHASE_IMAGE_TYPE},
                     is_regex=True),
        InputFilesets('gre_mag', GRE_PATTERN, dicom_format,
                        dicom_tags={IMAGE_TYPE_TAG: MAG_IMAGE_TYPE},
                        is_regex=True)]

    INPUTS_FROM_REF_DIR = True
    REF_FORMATS = [dicom_format]

    def test_dicom_match(self):
        study = self.create_study(
            TestMatchStudy, 'test_dicom',
            inputs=self.DICOM_MATCH)
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')

    def test_order_match(self):
        study = self.create_study(
            TestMatchStudy, 'test_dicom',
            inputs=[
                InputFilesets('gre_phase', pattern=self.GRE_PATTERN,
                              valid_formats=dicom_format, order=1,
                              is_regex=True),
                InputFilesets('gre_mag', pattern=self.GRE_PATTERN,
                              valid_formats=dicom_format, order=0,
                              is_regex=True)])
        phase = list(study.data('gre_phase'))[0]
        mag = list(study.data('gre_mag'))[0]
        self.assertEqual(phase.name, 'gre_field_mapping_3mm_phase')
        self.assertEqual(mag.name, 'gre_field_mapping_3mm_mag')


class TestDerivableStudy(with_metaclass(StudyMetaClass, Study)):

    add_data_specs = [
        InputFilesetSpec('required', text_format),
        InputFilesetSpec('optional', text_format, optional=True),
        FilesetSpec('derivable', text_format, 'pipeline1'),
        FilesetSpec('missing_input', text_format, 'pipeline2'),
        FilesetSpec('another_derivable', text_format, 'pipeline3'),
        FilesetSpec('requires_switch', text_format, 'pipeline3'),
        FilesetSpec('requires_switch2', text_format, 'pipeline4'),
        FilesetSpec('requires_foo', text_format, 'pipeline5'),
        FilesetSpec('requires_bar', text_format, 'pipeline5')]

    add_param_specs = [
        SwitchSpec('switch', False),
        SwitchSpec('branch', 'foo', ('foo', 'bar', 'wee'))]

    def pipeline1(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline1',
            desc="",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a']))
        pipeline.connect_input('required', identity, 'a')
        pipeline.connect_output('derivable', identity, 'a')
        return pipeline

    def pipeline2(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline2',
            desc="",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a', 'b']))
        pipeline.connect_input('required', identity, 'a')
        pipeline.connect_input('optional', identity, 'b')
        pipeline.connect_output('missing_input', identity, 'a')
        return pipeline

    def pipeline3(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline3',
            desc="",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a', 'b']))
        pipeline.connect_input('required', identity, 'a')
        pipeline.connect_input('required', identity, 'b')
        pipeline.connect_output('another_derivable', identity, 'a')
        if self.branch('switch'):
            pipeline.connect_output('requires_switch', identity, 'b')
        return pipeline

    def pipeline4(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline4',
            desc="",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a']))
        pipeline.connect_input('requires_switch', identity, 'a')
        pipeline.connect_output('requires_switch2', identity, 'a')
        return pipeline

    def pipeline5(self, **name_maps):
        pipeline = self.new_pipeline(
            'pipeline5',
            desc="",
            citations=[],
            name_maps=name_maps)
        identity = pipeline.add('identity', IdentityInterface(['a']))
        pipeline.connect_input('required', identity, 'a')
        if self.branch('branch', 'foo'):
            pipeline.connect_output('requires_foo', identity, 'a')
        elif self.branch('branch', 'bar'):
            pipeline.connect_output('requires_bar', identity, 'a')
        else:
            self.unhandled_branch('branch')
        return pipeline


class TestDerivable(BaseTestCase):

    INPUT_FILESETS = {'required': 'blah'}

    def test_derivable(self):
        # Test vanilla study
        study = self.create_study(
            TestDerivableStudy,
            'study',
            inputs={'required': 'required'})
        self.assertTrue(study.spec('derivable').derivable)
        self.assertTrue(
            study.spec('another_derivable').derivable)
        self.assertFalse(
            study.spec('missing_input').derivable)
        self.assertFalse(
            study.spec('requires_switch').derivable)
        self.assertFalse(
            study.spec('requires_switch2').derivable)
        self.assertTrue(study.spec('requires_foo').derivable)
        self.assertFalse(study.spec('requires_bar').derivable)
        # Test study with 'switch' enabled
        study_with_switch = self.create_study(
            TestDerivableStudy,
            'study_with_switch',
            inputs=[InputFilesets('required', 'required', text_format)],
            parameters={'switch': True})
        self.assertTrue(
            study_with_switch.spec('requires_switch').derivable)
        self.assertTrue(
            study_with_switch.spec('requires_switch2').derivable)
        # Test study with branch=='bar'
        study_bar_branch = self.create_study(
            TestDerivableStudy,
            'study_bar_branch',
            inputs=[InputFilesets('required', 'required', text_format)],
            parameters={'branch': 'bar'})
        self.assertFalse(study_bar_branch.spec('requires_foo').derivable)
        self.assertTrue(study_bar_branch.spec('requires_bar').derivable)
        # Test study with optional input
        study_with_input = self.create_study(
            TestDerivableStudy,
            'study_with_inputs',
            inputs=[InputFilesets('required', 'required', text_format),
                    InputFilesets('optional', 'required', text_format)])
        self.assertTrue(
            study_with_input.spec('missing_input').derivable)
        study_unhandled = self.create_study(
            TestDerivableStudy,
            'study_unhandled',
            inputs=[InputFilesets('required', 'required', text_format)],
            parameters={'branch': 'wee'})
        self.assertRaises(
            ArcanaDesignError,
            getattr,
            study_unhandled.spec('requires_foo'),
            'derivable')



from builtins import object
import os  # @UnusedImport
from pprint import pformat
import os.path as op
from collections import defaultdict, OrderedDict
import shutil
from itertools import repeat
from copy import copy, deepcopy
from logging import getLogger
import numpy as np
from nipype.pipeline import engine as pe
from nipype.interfaces.utility import IdentityInterface, Merge
from arcana.repository.interfaces import RepositorySource, RepositorySink
from arcana.utils import get_class_info
from arcana.exceptions import (
    ArcanaMissingDataException,
    ArcanaNoRunRequiredException, ArcanaUsageError, ArcanaDesignError,
    ArcanaReprocessException, ArcanaProtectedOutputConflictError,
    ArcanaOutputNotProducedException, ArcanaDataNotDerivedYetError,
    ArcanaNameError)


logger = getLogger('arcana')


class Processor(object):
    """
    A thin wrapper around the NiPype LinearPlugin used to connect
    runs pipelines on the local workstation

    Parameters
    ----------
    work_dir : str
        A directory in which to run the nipype workflows
    max_process_time : float
        The maximum time allowed for the process
    reprocess: bool | 'force'
        A flag which determines whether to rerun the processing for this
        step if a provenance mismatch is detected between save derivative and
        parameters passed to the Study. If False, an exception will be raised
        in this case. If passed a value of 'force' then the outputs will be
        regenerated regardless of existing derivatives (used mainly in testing)
    prov_check : iterable[str]
        Paths in the provenance dictionary to include in checks with previously
        generated derivatives to determine whether they need to be rerun.
        Paths are strings delimited by '/', with each part referring to a
        dictionary key in the nested provenance dictionary.
    prov_ignore : iterable[str]
        Paths in the provenance dictionary to exclude (if they are already
        included given the 'prov_check' kwarg) in checks with previously
        generated derivatives to determine whether they need to be rerun
        Paths are strings delimited by '/', with each part referring to a
        dictionary key in the nested provenance dictionary.
    clean_work_dir_between_runs : bool
        Whether to clean the working directory between runs (can avoid problems
        if debugging the analysis but may take longer to reach the same point)
    default_wall_time : int
        The default wall time assumed for nodes where it isn't specified
    default_mem_gb : float
        The default memory assumed to be required for nodes where it isn't
        specified
    """

    DEFAULT_WALL_TIME = 20
    DEFAULT_MEM_GB = 4

    WORKFLOW_MAX_NAME_LEN = 100

    # The default paths in the provenance JSON to check for mismatches that
    # would require the derivative to be reprocessed
    DEFAULT_PROV_CHECK = ['workflow', 'inputs', 'outputs', 'joined_ids']
    # The default paths in the provenance JSON to ignore mismatches that would
    # otherwise require the derivative to be reprocessed
    DEFAULT_PROV_IGNORE = ['.*/pkg_version',
                           'workflow/nodes/.*/requirements/.*/local_version',
                           'workflow/nodes/.*/requirements/.*/local_name']

    default_plugin_args = {}

    def __init__(self, work_dir, reprocess=False,
                 prov_check=DEFAULT_PROV_CHECK,
                 prov_ignore=DEFAULT_PROV_IGNORE,
                 max_process_time=None,
                 clean_work_dir_between_runs=True,
                 default_wall_time=DEFAULT_WALL_TIME,
                 default_mem_gb=DEFAULT_MEM_GB, **kwargs):
        self._work_dir = work_dir
        self._max_process_time = max_process_time
        self._reprocess = reprocess
        self._prov_check = prov_check
        self._prov_ignore = prov_ignore
        self._plugin_args = copy(self.default_plugin_args)
        self._default_wall_time = default_wall_time
        self._deffault_mem_gb = default_mem_gb
        self._plugin_args.update(kwargs)
        self._init_plugin()
        self._study = None
        self._clean_work_dir_between_runs = clean_work_dir_between_runs

    def __repr__(self):
        return "{}(work_dir={})".format(
            type(self).__name__, self._work_dir)

    def __eq__(self, other):
        try:
            return (
                self._work_dir == other._work_dir and
                self._max_process_time == other._max_process_time and
                self._reprocess == other._reprocess and
                self._plugin_args == other._plugin_args)
        except AttributeError:
            return False

    def _init_plugin(self):
        self._plugin = self.nipype_plugin_cls(**self._plugin_args)

    @property
    def study(self):
        return self._study

    @property
    def reprocess(self):
        return self._reprocess

    @property
    def prov_check(self):
        return self._prov_check

    @property
    def prov_ignore(self):
        return self._prov_ignore

    @property
    def default_mem_gb(self):
        return self._deffault_mem_gb

    @property
    def default_wall_time(self):
        return self._default_wall_time

    def bind(self, study):
        cpy = deepcopy(self)
        cpy._study = study
        return cpy

    def run(self, *pipelines, **kwargs):
        """
        Connects all pipelines to that study's repository and runs them
        in the same NiPype workflow

        Parameters
        ----------
        pipeline(s) : Pipeline, ...
            The pipeline to connect to repository
        required_outputs : list[set[str]]
            A set of required outputs for each pipeline. If None then all
            outputs are assumed to be required
        subject_ids : list[str]
            The subset of subject IDs to process. If None all available will be
            processed. Note this is not a duplication of the study
            and visit IDs passed to the Study __init__, as they define the
            scope of the analysis and these simply limit the scope of the
            current run (e.g. to break the analysis into smaller chunks and
            run separately). Therefore, if the analysis joins over subjects,
            then all subjects will be processed and this parameter will be
            ignored.
        visit_ids : list[str]
            The same as 'subject_ids' but for visit IDs
        session_ids : list[str,str]
            The same as 'subject_ids' and 'visit_ids', except specifies a set
            of specific combinations in tuples of (subject ID, visit ID).
        force : bool | 'all'
            A flag to force the reprocessing of all sessions in the filter
            array, regardless of whether the parameters|pipeline used
            to generate them matches the current ones. NB: if True only the
            final pipeline will be reprocessed (prerequisite pipelines won't
            run unless they don't match provenance). To process all
            prerequisite pipelines 'all' should be passed to force.

        Returns
        -------
        report : ReportNode
            The final report node, which can be connected to subsequent
            pipelines
        """
        if not pipelines:
            raise ArcanaUsageError("No pipelines provided to {}.run"
                                   .format(self))
        # Get filter kwargs  (NB: in Python 3 they could be in the kwarg list)
        subject_ids = kwargs.pop('subject_ids', None)
        visit_ids = kwargs.pop('visit_ids', None)
        session_ids = kwargs.pop('session_ids', None)
        clean_work_dir = kwargs.pop('clean_work_dir',
                                    self._clean_work_dir_between_runs)
        required_outputs = kwargs.pop('required_outputs', repeat(None))
        # Create name by combining pipelines
        name = '_'.join(p.name for p in pipelines)
        # Clean work dir if required
        if clean_work_dir:
            workflow_work_dir = op.join(self.work_dir, name)
            if op.exists(workflow_work_dir):
                shutil.rmtree(workflow_work_dir)
        # Trim the end of very large names to avoid problems with
        # workflow names exceeding system limits.
        name = name[:self.WORKFLOW_MAX_NAME_LEN]
        workflow = pe.Workflow(name=name, base_dir=self.work_dir)

        # Generate filter array to optionally restrict the run to certain
        # subject and visit IDs.
        tree = self.study.tree
        # Create maps from the subject|visit IDs to an index used to represent
        # them in the filter array
        subject_inds = {s.id: i for i, s in enumerate(tree.subjects)}
        visit_inds = {v.id: i for i, v in enumerate(tree.visits)}
        if not subject_ids and not visit_ids and not session_ids:
            # No filters applied so create a full filter array
            filter_array = np.ones((len(subject_inds), len(visit_inds)),
                                   dtype=bool)
        else:
            # Filters applied so create an empty filter array and populate
            # from filter lists
            filter_array = np.zeros((len(subject_inds), len(visit_inds)),
                                    dtype=bool)
            if subject_ids is not None:
                for subj_id in subject_ids:
                    filter_array[subject_inds[subj_id], :] = True
            if visit_ids is not None:
                for visit_id in visit_ids:
                    filter_array[:, visit_inds[visit_id]] = True
            if session_ids is not None:
                for subj_id, visit_id in session_ids:
                    filter_array[subject_inds[subj_id],
                                 visit_inds[visit_id]] = True
            if not filter_array.any():
                raise ArcanaUsageError(
                    "Provided filters:\n" +
                    ("  subject_ids: {}\n".format(', '.join(subject_ids))
                     if subject_ids is not None else '') +
                    ("  visit_ids: {}\n".format(', '.join(visit_ids))
                     if visit_ids is not None else '') +
                    ("  session_ids: {}\n".format(', '.join(session_ids))
                     if session_ids is not None else '') +
                    "Did not match any sessions in the project:\n" +
                    "  subject_ids: {}\n".format(', '.join(subject_inds)) +
                    "  visit_ids: {}\n".format(', '.join(visit_inds)))

        # Stack of pipelines to process in reverse order of required execution
        stack = OrderedDict()

        def push_on_stack(pipeline, filt_array, req_outputs, downstream=()):
            """
            Push a pipeline onto the stack of pipelines to be processed,
            detecting common upstream pipelines and resolving them to a single
            pipeline

            Parameters
            ----------
            pipeline : Pipeline
                The pipeline to add to the stack
            filt_array : 2-D numpy.ndarray
                An array of sessions that are to be processed for the given
                pipeline
            req_outputs : list[str]
                The outputs of the pipeline that are actually required (non-
                required outputs can be ignored if they are not generated)
            downstream : tuple[Pipeline]
                The pipelines directly downstream of the pipeline to be added.
                Used to detect circular dependencies
            """
            if req_outputs is None:
                req_outputs = pipeline.output_names
            # Check downstream piplines for circular dependencies
            downstream_pipelines = [p for p, _ in downstream]
            if pipeline in downstream_pipelines:
                recur_index = downstream_pipelines.index(pipeline)
                raise ArcanaDesignError(
                    "{} cannot be a dependency of itself. Call-stack:\n{}"
                    .format(pipeline,
                            '\n'.join('{} ({})'.format(p, ', '.join(ro))
                                      for p, ro in (
                                          ((pipeline, req_outputs),) +
                                          downstream[:(recur_index + 1)]))))
            if pipeline.name in stack:
                # Pop pipeline from stack in order to add it to the end of the
                # stack and ensure it is run before all downstream pipelines
                prev_pipeline, prev_req_outputs, prev_filt_array = stack.pop(
                    pipeline.name)
                if pipeline is not prev_pipeline and pipeline != prev_pipeline:
                    raise ArcanaDesignError(
                        "Attempting to run two different pipelines with the "
                        "same name, {} and {}".format(pipeline, prev_pipeline))
                # Combined required outputs and filter array with previous
                # references to the pipeline
                req_outputs = copy(req_outputs)
                req_outputs.update(prev_req_outputs)
                filt_array = filt_array | prev_filt_array
            # If the pipeline to process contains summary outputs (i.e. 'per-
            # subject|visit|study' frequency), then we need to "dialate" the
            # filter array to include IDs across the scope of the study, e.g.
            # all subjects for per-vist, or all visits for per-subject.
            output_freqs = set(pipeline.output_frequencies)
            dialated_filt_array = self._dialate_array(filt_array,
                                                      pipeline.joins)
            added = dialated_filt_array ^ filt_array
            if added.any():
                filt_array = dialated_filt_array
                # Invert the index dictionaries to get index-to-ID maps
                inv_subject_inds = {v: k for k, v in subject_inds.items()}
                inv_visit_inds = {v: k for k, v in visit_inds.items()}
                logger.warning(
                    "Dialated filter array used to process '{}' pipeline to "
                    "include {} subject/visit IDs due to its '{}' summary "
                    "outputs ".format(
                        pipeline.name,
                        ', '.join('({},{})'.format(inv_subject_inds[s],
                                                   inv_visit_inds[v])
                                  for s, v in zip(*np.nonzero(added))),
                        "' and '".join(output_freqs)))

            stack[pipeline.name] = pipeline, req_outputs, filt_array
            # Recursively add all prerequisites to stack
            try:
                for (prq_getter,
                     prq_req_outputs) in pipeline.prerequisites.items():
                    prereq = pipeline.study.pipeline(
                        prq_getter, prq_req_outputs)
                    push_on_stack(prereq, filt_array, prq_req_outputs,
                                  ((pipeline, req_outputs),) + downstream)
            except (ArcanaMissingDataException,
                    ArcanaOutputNotProducedException) as e:
                e.msg += ("\nwhich are required as inputs to the '{}' "
                          "pipeline to produce '{}'".format(
                              pipeline.name, "', '".join(req_outputs)))
                raise e

        # Add all primary pipelines to the stack along with their prereqs
        for pipeline, req_outputs in zip(pipelines, required_outputs):
            push_on_stack(pipeline, filter_array, req_outputs)

        # Iterate through stack of required pipelines from upstream to
        # downstream
        for pipeline, req_outputs, flt_array in reversed(list(stack.values())):
            try:
                self._connect_pipeline(
                    pipeline, req_outputs, workflow, subject_inds, visit_inds,
                    flt_array, **kwargs)
            except ArcanaNoRunRequiredException:
                logger.info("Not running '{}' pipeline as its outputs "
                            "are already present in the repository"
                            .format(pipeline.name))
        # Save complete graph for debugging purposes
#         workflow.write_graph(graph2use='flat', format='svg')
#         print('Graph saved in {} directory'.format(os.getcwd()))
        # Actually run the generated workflow
        if workflow._get_all_nodes():  # Check if workflow has any nodes to run
            result = workflow.run(plugin=self._plugin)
        else:
            result = None
        # Reset the cached tree of filesets in the repository as it will
        # change after the pipeline has run.
        self.study.clear_caches()
        return result

    def _connect_pipeline(self, pipeline, required_outputs, workflow,
                          subject_inds, visit_inds, filter_array, force=False):
        """
        Connects a pipeline to a overarching workflow that sets up iterators
        over subjects|visits present in the repository (if required) and
        repository source and sink nodes

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline to connect
        required_outputs : set[str] | None
            The outputs required to be produced by this pipeline. If None all
            are deemed to be required
        workflow : nipype.pipeline.engine.Workflow
            The overarching workflow to connect the pipeline to
        subject_inds : dct[str, int]
            A mapping of subject ID to row index in the filter array
        visit_inds : dct[str, int]
            A mapping of visit ID to column index in the filter array
        filter_array : 2-D numpy.array[bool]
            A two-dimensional boolean array, where rows correspond to
            subjects and columns correspond to visits in the repository. True
            values represent a combination of subject & visit ID to include
            in the current round of processing. Note that if the 'force'
            flag is not set, sessions won't be reprocessed unless the
            save provenance doesn't match that of the given pipeline.
        force : bool | 'all'
            A flag to force the processing of all sessions in the filter
            array, regardless of whether the parameters|pipeline used
            to generate existing data matches the given pipeline
        """
        if self.reprocess == 'force':
            force = True
        # Close-off construction of the pipeline and created, input and output
        # nodes and provenance dictionary
        pipeline.cap()
        # Prepend prerequisite pipelines to complete workflow if they need
        # to be (re)processed
        final_nodes = []
        # The array that represents the subject/visit pairs for which any
        # prerequisite pipeline will be (re)processed, and which therefore
        # needs to be included in the processing of the current pipeline. Row
        # indices correspond to subjects and column indices visits
        prqs_to_process_array = np.zeros((len(subject_inds), len(visit_inds)),
                                         dtype=bool)
        # The array that represents the subject/visit pairs for which any
        # prerequisite pipeline will be skipped due to missing inputs. Row
        # indices correspond to subjects and column indices visits
        prqs_to_skip_array = np.zeros((len(subject_inds), len(visit_inds)),
                                      dtype=bool)
        for getter_name in pipeline.prerequisites:
            prereq = pipeline.study.pipeline(getter_name)
            if prereq.to_process_array.any():
                final_nodes.append(prereq.node('final'))
                prqs_to_process_array |= prereq.to_process_array
            prqs_to_skip_array |= prereq.to_skip_array
        # Get list of sessions that need to be processed (i.e. if
        # they don't contain the outputs of this pipeline)
        to_process_array, to_protect_array, to_skip_array = self._to_process(
            pipeline, required_outputs, prqs_to_process_array,
            prqs_to_skip_array, filter_array, subject_inds, visit_inds, force)
        # Store the arrays signifying which nodes to process, protect or skip
        # so they can be passed to downstream pipelines
        pipeline.to_process_array = to_process_array
        pipeline.to_protect_array = to_protect_array
        pipeline.to_skip_array = to_skip_array
        # Check to see if there are any sessions to process
        if not to_process_array.any():
            raise ArcanaNoRunRequiredException(
                "No sessions to process for '{}' pipeline"
                .format(pipeline.name))
        # Set up workflow to run the pipeline, loading and saving from the
        # repository
        workflow.add_nodes([pipeline._workflow])
        # If prerequisite pipelines need to be processed, connect their
        # "final" nodes to the initial node of this pipeline to ensure that
        # they are all processed before this pipeline is run.
        if final_nodes:
            prereqs = pipeline.add('prereqs', Merge(len(final_nodes)))
            for i, final_node in enumerate(final_nodes, start=1):
                workflow.connect(final_node, 'out', prereqs, 'in{}'.format(i))
        else:
            prereqs = None
        # Construct iterator structure over subjects and sessions to be
        # processed
        iter_nodes = self._iterate(pipeline, to_process_array, subject_inds,
                                   visit_inds)
        sources = {}
        # Loop through each frequency present in the pipeline inputs and
        # create a corresponding source node
        for freq in pipeline.input_frequencies:
            try:
                inputs = list(pipeline.frequency_inputs(freq))
            except ArcanaMissingDataException as e:
                raise ArcanaMissingDataException(
                    str(e) + ", which is required for pipeline '{}'".format(
                        pipeline.name))
            inputnode = pipeline.inputnode(freq)
            sources[freq] = source = pipeline.add(
                '{}_source'.format(freq),
                RepositorySource(
                    i.collection for i in inputs),
                inputs=({'prereqs': (prereqs, 'out')} if prereqs is not None
                         else {}))
            # Connect iter_nodes to source and input nodes
            for iterator in pipeline.iterators(freq):
                pipeline.connect(iter_nodes[iterator], iterator, source,
                                 iterator)
                pipeline.connect(source, iterator, inputnode,
                                 iterator)
            for input in inputs:  # @ReservedAssignment
                pipeline.connect(source, input.suffixed_name,
                                 inputnode, input.name)
        deiter_nodes = {}

        def deiter_node_sort_key(it):
            """
            If there are two iter_nodes (i.e. both subject and visit ID) and
            one depends on the other (i.e. if the visit IDs per subject
            vary and vice-versa) we need to ensure that the dependent
            iterator is deiterated (joined) first.
            """
            return iter_nodes[it].itersource is None

        # Connect all outputs to the repository sink, creating a new sink for
        # each frequency level (i.e 'per_session', 'per_subject', 'per_visit',
        # or 'per_study')
        for freq in pipeline.output_frequencies:
            outputs = list(pipeline.frequency_outputs(freq))
            if pipeline.iterators(freq) - pipeline.iterators():
                raise ArcanaDesignError(
                    "Doesn't make sense to output '{}', which are of '{}' "
                    "frequency, when the pipeline only iterates over '{}'"
                    .format("', '".join(o.name for o in outputs), freq,
                            "', '".join(pipeline.iterators())))
            outputnode = pipeline.outputnode(freq)
            # Connect filesets/fields to sink to sink node, skipping outputs
            # that are study inputs
            to_connect = {o.suffixed_name: (outputnode, o.name)
                          for o in outputs if o.is_spec}
            # Connect iterators to sink node
            to_connect.update(
                {i: (iter_nodes[i], i) for i in pipeline.iterators()})
            # Connect checksums/values from sources to sink node in order to
            # save in provenance, joining where necessary
            for input_freq in pipeline.input_frequencies:
                checksums_to_connect = [
                    i.checksum_suffixed_name
                    for i in pipeline.frequency_inputs(input_freq)]
                if not checksums_to_connect:
                    # Rare case of a pipeline with no inputs only iter_nodes
                    # that will only occur in unittests in all likelihood
                    continue
                # Loop over iterators that need to be joined, i.e. that are
                # present in the input frequency but not the output frequency
                # and create join nodes
                source = sources[input_freq]
                for iterator in (pipeline.iterators(input_freq) -
                                  pipeline.iterators(freq)):
                    join = pipeline.add(
                        '{}_to_{}_{}_checksum_join'.format(
                            input_freq, freq, iterator),
                        IdentityInterface(
                            checksums_to_connect),
                        inputs={
                            tc: (source, tc) for tc in checksums_to_connect},
                        joinsource=iterator,
                        joinfield=checksums_to_connect)
                    source = join
                to_connect.update(
                    {tc: (source, tc) for tc in checksums_to_connect})
            # Add sink node
            sink = pipeline.add(
                '{}_sink'.format(freq),
                RepositorySink(
                    (o.collection for o in outputs), pipeline,
                    required_outputs),
                inputs=to_connect)
            # "De-iterate" (join) over iterators to get back to single child
            # node by the time we connect to the final node of the pipeline Set
            # the sink and subject_id as the default deiterator if there are no
            # deiterates (i.e. per_study) or to use as the upstream node to
            # connect the first deiterator for every frequency
            deiter_nodes[freq] = sink  # for per_study the "deiterator" == sink
            for iterator in sorted(pipeline.iterators(freq),
                                    key=deiter_node_sort_key):
                # Connect to previous deiterator or sink
                # NB: we only need to keep a reference to the last one in the
                # chain in order to connect with the "final" node, so we can
                # overwrite the entry in the 'deiter_nodes' dict
                deiter_nodes[freq] = pipeline.add(
                    '{}_{}_deiter'.format(freq, iterator),
                    IdentityInterface(
                        ['checksums']),
                    inputs={
                        'checksums': (deiter_nodes[freq], 'checksums')},
                    joinsource=iterator,
                    joinfield='checksums')
        # Create a final node, which is used to connect with downstream
        # pipelines
        pipeline.add(
            'final',
            Merge(
                len(deiter_nodes)),
            inputs={
                'in{}'.format(i): (di, 'checksums')
                for i, di in enumerate(deiter_nodes.values(), start=1)})

    def _iterate(self, pipeline, to_process_array, subject_inds, visit_inds):
        """
        Generate nodes that iterate over subjects and visits in the study that
        need to be processed by the pipeline

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline to add iter_nodes for
        to_process_array : 2-D numpy.array[bool]
            A two-dimensional boolean array, where rows correspond to
            subjects and columns correspond to visits in the repository. True
            values represent a combination of subject & visit ID to process
            the session for
        subject_inds : dct[str, int]
            A mapping of subject ID to row index in the 'to_process' array
        visit_inds : dct[str, int]
            A mapping of visit ID to column index in the 'to_process' array

        Returns
        -------
        iter_nodes : dict[str, Node]
            A dictionary containing the nodes to iterate over all subject/visit
            IDs to process for each input frequency
        """
        # Check to see whether the subject/visit IDs to process (as specified
        # by the 'to_process' array) can be factorized into indepdent nodes,
        # i.e. all subjects to process have the same visits to process and
        # vice-versa.
        factorizable = True
        if len(list(pipeline.iterators())) == 2:
            nz_rows = to_process_array[to_process_array.any(axis=1), :]
            ref_row = nz_rows[0, :]
            factorizable = all((r == ref_row).all() for r in nz_rows)
        # If the subject/visit IDs to process cannot be factorized into
        # indepedent iterators, determine which to make make dependent on the
        # other in order to avoid/minimise duplicatation of download attempts
        dependent = None
        if not factorizable:
            input_freqs = list(pipeline.input_frequencies)
            # By default pick iterator the one with the most IDs to
            # iterate to be the dependent in order to reduce the number of
            # nodes created and any duplication of download attempts across
            # the nodes (if both 'per_visit' and 'per_subject' inputs are
            # required
            num_subjs, num_visits = nz_rows[:, nz_rows.any(axis=0)].shape
            if num_subjs > num_visits:
                dependent = self.study.SUBJECT_ID
            else:
                dependent = self.study.VISIT_ID
            if 'per_visit' in input_freqs:
                if 'per_subject' in input_freqs:
                    logger.warning(
                        "Cannot factorize sessions to process into independent"
                        " subject and visit iterators and both 'per_visit' and"
                        " 'per_subject' inputs are used by pipeline therefore"
                        " per_{} inputs may be cached twice".format(
                            dependent[:-3]))
                else:
                    dependent = self.study.SUBJECT_ID
            elif 'per_subject' in input_freqs:
                dependent = self.study.VISIT_ID
        # Invert the index dictionaries to get index-to-ID maps
        inv_subj_inds = {v: k for k, v in subject_inds.items()}
        inv_visit_inds = {v: k for k, v in visit_inds.items()}
        # Create iterator for subjects
        iter_nodes = {}
        if self.study.SUBJECT_ID in pipeline.iterators():
            fields = [self.study.SUBJECT_ID]
            if dependent == self.study.SUBJECT_ID:
                fields.append(self.study.VISIT_ID)
            # Add iterator node named after subject iterator
            subj_it = pipeline.add(self.study.SUBJECT_ID,
                                   IdentityInterface(fields))
            if dependent == self.study.SUBJECT_ID:
                # Subjects iterator is dependent on visit iterator (because of
                # non-factorizable IDs)
                subj_it.itersource = ('{}_{}'.format(pipeline.name,
                                                     self.study.VISIT_ID),
                                      self.study.VISIT_ID)
                subj_it.iterables = [(
                    self.study.SUBJECT_ID,
                    {inv_visit_inds[n]: [inv_subj_inds[m]
                                         for m in col.nonzero()[0]]
                     for n, col in enumerate(to_process_array.T)})]
            else:
                subj_it.iterables = (
                    self.study.SUBJECT_ID,
                    [inv_subj_inds[n]
                     for n in to_process_array.any(axis=1).nonzero()[0]])
            iter_nodes[self.study.SUBJECT_ID] = subj_it
        # Create iterator for visits
        if self.study.VISIT_ID in pipeline.iterators():
            fields = [self.study.VISIT_ID]
            if dependent == self.study.VISIT_ID:
                fields.append(self.study.SUBJECT_ID)
            # Add iterator node named after visit iterator
            visit_it = pipeline.add(self.study.VISIT_ID,
                                    IdentityInterface(fields))
            if dependent == self.study.VISIT_ID:
                visit_it.itersource = ('{}_{}'.format(pipeline.name,
                                                      self.study.SUBJECT_ID),
                                       self.study.SUBJECT_ID)
                visit_it.iterables = [(
                    self.study.VISIT_ID,
                    {inv_subj_inds[m]:[inv_visit_inds[n]
                                       for n in row.nonzero()[0]]
                     for m, row in enumerate(to_process_array)})]
            else:
                visit_it.iterables = (
                    self.study.VISIT_ID,
                    [inv_visit_inds[n]
                     for n in to_process_array.any(axis=0).nonzero()[0]])
            iter_nodes[self.study.VISIT_ID] = visit_it
        if dependent == self.study.SUBJECT_ID:
            pipeline.connect(visit_it, self.study.VISIT_ID,
                             subj_it, self.study.VISIT_ID)
        if dependent == self.study.VISIT_ID:
            pipeline.connect(subj_it, self.study.SUBJECT_ID,
                             visit_it, self.study.SUBJECT_ID)
        return iter_nodes

    def _to_process(self, pipeline, required_outputs, prqs_to_process_array,
                    to_skip_array, filter_array, subject_inds, visit_inds,
                    force):
        """
        Check whether the outputs of the pipeline are present in all sessions
        in the project repository and were generated with matching provenance.
        Return an 2D boolean array (subjects: rows, visits: cols) with the
        sessions to process marked True.

        Parameters
        ----------
        pipeline : Pipeline
            The pipeline to determine the sessions to process
        required_ouputs : set[str]
            The names of the pipeline outputs that are required. If None all
            are deemed to be required
        prqs_to_process_array : 2-D numpy.array[bool]
            A two-dimensional boolean array, where rows and columns correspond
            correspond to subjects and visits in the repository tree. True
            values represent a subject/visit ID pairs that will be
            (re)processed in prerequisite pipelines and therefore need to be
            included in the returned array.
        to_skip_array : 2-D numpy.array[bool]
            Similar to prqs_to_process_array, except denote the subject/visits
            that are to be skipped due to missing inputs
        filter_array : 2-D numpy.array[bool]
            A two-dimensional boolean array, where rows and columns correspond
            correspond to subjects and visits in the repository tree. True
            values represent a subject/visit ID pairs to include
            in the current round of processing. Note that if the 'force'
            flag is not set, sessions won't be reprocessed unless the
            parameters and pipeline version saved in the provenance doesn't
            match that of the given pipeline.
        subject_inds : dict[str,int]
            Mapping from subject ID to index in filter|to_process arrays
        visit_inds : dict[str,int]
            Mapping from visit ID to index in filter|to_process arrays
        force : bool
            Whether to force reprocessing of all (filtered) sessions or not.
            Note that if 'force' is true we can't just return the filter array
            as it might be dilated by summary outputs (i.e. of frequency
            'per_visit', 'per_subject' or 'per_study'). So we still loop
            through all outputs and treat them like they don't exist

        Returns
        -------
        to_process_array : 2-D numpy.array[bool]
            A two-dimensional boolean array, where rows correspond to
            subjects and columns correspond to visits in the repository. True
            values represent subject/visit ID pairs to run the pipeline for
        """
        # Reference the study tree in local variable for convenience
        tree = self.study.tree
        # Check to see if the pipeline has any low frequency outputs, because
        # if not then each session can be processed indepdently. Otherwise,
        # the "session matrix" (as defined by subject_ids and visit_ids
        # passed to the Study class) needs to be complete, i.e. a session
        # exists (with the full complement of requird inputs) for each
        # subject/visit ID pair.
        summary_outputs = [
            o.name for o in pipeline.outputs if o.frequency != 'per_session']
        # Set of frequencies present in pipeline outputs
        output_freqs = pipeline.output_frequencies
        if summary_outputs:
            if list(tree.incomplete_subjects):
                raise ArcanaUsageError(
                    "Can't process '{}' pipeline as it has low frequency "
                    " outputs (i.e. outputs that aren't of 'per_session' "
                    "frequency) ({}) and subjects ({}) that are missing one "
                    "or more visits ({}). Please restrict the subject/visit "
                    "IDs in the study __init__ to continue the analysis"
                    .format(
                        pipeline.name,
                        ', '.join(summary_outputs),
                        ', '.join(s.id for s in tree.incomplete_subjects),
                        ', '.join(v.id for v in tree.incomplete_visits)))

        def array_inds(x):
            """
            Get the array index corresponding to an item or node. If it doesn't
            have a subject and/or visit ID (i.e. frequency != 'per_session')
            then the first index in the row/column is selected as it will be
            dialated to the whole row/column later
            """
            return (subject_inds.get(x.subject_id, 0),
                    visit_inds.get(x.visit_id, 0))

        # Initalise array to represent which sessions need to be reprocessed
        to_process_array = np.zeros((len(subject_inds), len(visit_inds)),
                                    dtype=bool)
        # An array to mark outputs that have been altered outside of Arcana
        # and therefore protect from over-writing
        to_protect_array = np.copy(to_process_array)
        # Mark the sessions that we should check to see if the configuration
        # saved in the provenance record matches that of the current pipeline
        to_check_array = np.copy(to_process_array)
        # As well as the the sessions that need to be protected or skipped,
        # keep track of the items in those sessions that are protected or
        # missing for more informative error messages
        to_protect = defaultdict(list)
        to_skip = defaultdict(list)
        # Check data tree for missing inputs
        for input in pipeline.inputs:  # @ReservedAssignment
            # NB: Study inputs that don't have skip_missing set and have
            # missing data should raise an error before this point
            if input.skip_missing:
                for item in input.collection:
                    if not item.exists:
                        to_skip_array[array_inds(item)] = True
                        to_skip[array_inds(item)].append(item)
        # Dialate array over all iterators that are joined by the pipeline
        to_skip_array = self._dialate_array(to_skip_array, pipeline.joins)
        # Check data tree for missing required outputs
        for output in pipeline.outputs:
            # Check to see if output is required by downstream processing
            required = (required_outputs is None or
                        output.name in required_outputs)
            for item in output.collection:
                if item.exists:
                    # Check to see if checksums recorded when derivative
                    # was generated by previous run match those of current file
                    # set. If not we assume they have been manually altered and
                    # therefore should not be overridden
                    if item.checksums != item.recorded_checksums:
                        logger.warning(
                            "Checksums for {} do not match those recorded in "
                            "provenance. Assuming it has been manually "
                            "corrected outside of Arcana and will therefore "
                            "not overwrite. Please delete manually if this "
                            "is not intended".format(repr(item)))
                        to_protect_array[array_inds(item)] = True
                        to_protect[array_inds(item)].append(item)
                    elif required:
                        if force:
                            to_process_array[array_inds(item)] = True
                        else:
                            to_check_array[array_inds(item)] = True
                elif required:
                    to_process_array[array_inds(item)] = True
        # Filter sessions to process by those requested
        to_process_array *= filter_array
        to_check_array *= (filter_array * np.invert(to_process_array))
        if to_check_array.any() and self.prov_check:
            # Get list of sessions, subjects, visits, tree objects to check
            # their provenance against that of the pipeline
            to_check = []
            if 'per_session' in output_freqs:
                to_check.extend(
                    s for s in tree.sessions
                    if to_check_array[subject_inds[s.subject_id],
                                      visit_inds[s.visit_id]])
            if 'per_subject' in output_freqs:
                # We can just test the first col of outputs_exist as rows
                # should be either all True or all False
                to_check.extend(s for s in tree.subjects
                                if to_check_array[subject_inds[s.id], 0])
            if 'per_visit' in output_freqs:
                # We can just test the first row of outputs_exist as cols
                # should be either all True or all False
                to_check.extend(v for v in tree.visits
                                if to_check_array[0, visit_inds[v.id]])
            if 'per_study' in output_freqs:
                to_check.append(tree)
            for node in to_check:
                # Generated expected record from current pipeline/repository-
                # state
                requires_reprocess = False
                try:
                    # Retrieve record stored in tree node
                    record = node.record(pipeline.name, pipeline.study.name)
                    expected_record = pipeline.expected_record(node)

                    # Compare record with expected
                    mismatches = record.mismatches(expected_record,
                                                   self.prov_check,
                                                   self.prov_ignore)
                    if mismatches:
                        msg = ("mismatch in provenance:\n{}\n Add mismatching "
                               "paths (delimeted by '/') to 'prov_ignore' "
                               "argument of Processor to ignore"
                               .format(
                                   pformat(mismatches)))
                        requires_reprocess = True
                except ArcanaNameError as e:
                    msg = "missing provenance record"
                    requires_reprocess = False
                    to_protect_array[array_inds(node)] = True
                except ArcanaDataNotDerivedYetError as e:
                    msg = ("missing input '{}' and therefore cannot check "
                           "provenance".format(e.name))
                    requires_reprocess = True
                if requires_reprocess:
                    if self.reprocess:
                        to_process_array[array_inds(node)] = True
                        logger.info(
                            "Reprocessing {} with '{}' due to {}"
                            .format(node, pipeline.name, msg))
                    else:
                        raise ArcanaReprocessException(
                            "Cannot use derivatives for '{}' pipeline stored "
                            "in {} due to {}, set 'reprocess' "
                            "flag of Processor to overwrite".format(
                                pipeline.name, node, msg))

        def inds_to_ids(inds):
            subject_id = next(k for k, v in subject_inds.items()
                              if v == inds[0])
            visit_id = next(k for k, v in visit_inds.items() if v == inds[1])
            return (subject_id, visit_id)

        # Dialate to process array
        to_process_array = self._dialate_array(to_process_array,
                                               pipeline.joins)
        intersection = to_process_array * to_skip_array
        if intersection.any():
            missing_inputs_msg = ''
            missing_prq_inputs_msg = ''
            for sess_inds in zip(*np.nonzero(intersection)):
                subject_id, visit_id = inds_to_ids(sess_inds)
                if sess_inds in to_skip:
                    missing_inputs_msg += (
                        "\n(subject={}, visit={}): [{}]".format(
                            subject_id, visit_id,
                            ', '.join(i.name for i in to_skip[sess_inds])))
                else:
                    missing_prq_inputs_msg += (
                        "\n(subject={}, visit={})".format(subject_id,
                                                          visit_id))
            warning_msg = ("Skipping the following nodes for '{}' pipeline due"
                           " to ".format(pipeline))
            if missing_inputs_msg:
                warning_msg += "missing inputs:{}".format(missing_inputs_msg)
            if missing_prq_inputs_msg:
                if missing_inputs_msg:
                    warning_msg += "\nand the following due to "
                warning_msg += ("missing inputs to prerequisite pipelines:{}"
                                .format(missing_prq_inputs_msg))
            logger.warning(warning_msg)
        # Remove nodes that are to be skipped due to missing inputs
        to_process_array *= np.invert(to_skip_array)
        # Check for conflicts between nodes to process and nodes to protect
        conflicting = to_process_array * to_protect_array
        if conflicting.any():
            error_msg = ''
            for sess_inds in zip(*np.nonzero(conflicting)):
                subject_id, visit_id = inds_to_ids(sess_inds)
                if required_outputs is None:
                    conflict_outputs = pipeline.outputs
                else:
                    conflict_outputs = [pipeline.study.bound_spec(r)
                                        for r in required_outputs]
                items = [
                    o.collection.item(subject_id=subject_id, visit_id=visit_id)
                    for o in conflict_outputs]
                missing = [i for i in items if i not in to_protect[sess_inds]]
                error_msg += (
                    "\n({}, {}): protected=[{}], missing=[{}]"
                    .format(
                        subject_id, visit_id,
                        ', '.join(i.name for i in to_protect[sess_inds]),
                        ', '.join(i.name for i in missing)))
            raise ArcanaProtectedOutputConflictError(
                "Cannot process {} as there are nodes with both protected "
                "outputs (ones modified externally to Arcana) and missing "
                "required outputs. Either delete protected outputs or provide "
                "missing required outputs to continue:{}".format(pipeline,
                                                                 error_msg))
        # Add in any prerequisites to process that aren't explicitly protected
        to_process_array |= (prqs_to_process_array *
                             filter_array *
                             np.invert(to_protect_array))
        to_process_array = self._dialate_array(to_process_array,
                                               pipeline.joins)
        return to_process_array, to_protect_array, to_skip_array

    def _dialate_array(self, array, iterators):
        """
        'Dialates' a to_process/to_protect array to include all subject and/or
        visits if the pipeline contains any joins over the corresponding
        iterators.

        Parameters
        ----------
        array : np.array[M, N]
            The array to potentially dialate
        iterators : set[str]
            The iterators that the array should be dialated for

        Returns
        -------
        dialated : np.array[M, N]
            The dialated array
        """
        if not iterators:
            return array
        dialated = np.copy(array)
        if self.study.SUBJECT_ID in iterators:
            # If we join over subjects we should include all subjects for every
            # visit we want to process
            dialated[:, dialated.any(axis=0)] = True
        if self.study.VISIT_ID in iterators:
            # If we join over visits we should include all visits for every
            # subject we want to process
            dialated[dialated.any(axis=1), :] = True
        return dialated

    @property
    def work_dir(self):
        return self._work_dir

    def __getstate__(self):
        dct = copy(self.__dict__)
        # Delete the NiPype plugin as it can be regenerated
        del dct['_plugin']
        return dct

    def __setstate__(self, state):
        self.__dict__.update(state)
        self._init_plugin()

    @property
    def prov(self):
        return {'type': get_class_info(type(self))}



from past.builtins import basestring
from builtins import object
from itertools import chain
from collections import defaultdict
import sys
import os.path as op
import types
from copy import copy
from logging import getLogger
from nipype.interfaces.utility import IdentityInterface
from arcana.exceptions import (
    ArcanaMissingInputError, ArcanaNoConverterError, ArcanaDesignError,
    ArcanaCantPickleStudyError, ArcanaUsageError,
    ArcanaMissingDataException, ArcanaNameError,
    ArcanaOutputNotProducedException, ArcanaInputMissingMatchError,
    ArcanaInputError)
from arcana.pipeline import Pipeline
from arcana.data import (
    BaseData, BaseInput, BaseInputSpec, InputFilesets, InputFields,
    BaseFileset)
from nipype.pipeline import engine as pe
from .parameter import Parameter, SwitchSpec
from arcana.repository.interfaces import RepositorySource
from arcana.repository import BasicRepo
from arcana.processor import SingleProc
from arcana.environment import StaticEnv
from arcana.utils import get_class_info

logger = getLogger('arcana')


class Study(object):
    """

    Parameters
    ----------
    name : str
        The name of the study.
    repository : Repository
        An Repository object that provides access to a DaRIS, XNAT or local file
        system
    processor : Processor
        A Processor to process the pipelines required to generate the
        requested derived filesets.
    inputs : Dict[str, InputFilesets | FilesetSpec | InputFields | FieldSpec] | List[InputFilesets | FilesetSpec | InputFields | FieldSpec]
        Either a list or a dictionary containing InputFilesets,
        InputFields, FilesetSpec, or FieldSpec objects, which specify the
        names of input filesets to the study, i.e. those that won't
        be generated by this study (although can be derived by the parent
        MultiStudy)
    environment : Environment
        An Environment within which to process the pipelines. Handles the
        version management + loading/unloading of software requirements
    parameters : List[Parameter] | Dict[str, (int|float|str)]
        Parameters that are passed to pipelines when they are constructed
        either as a dictionary of key-value pairs or as a list of
        'Parameter' objects. The name and dtype must match ParamSpecs in
        the _param_spec class attribute (see 'add_param_specs').
    subject_ids : List[(int|str)]
        List of subject IDs to restrict the analysis to
    visit_ids : List[(int|str)]
        List of visit IDs to restrict the analysis to
    enforce_inputs : bool (default: True)
        Whether to check the inputs to see if any acquired filesets
        are missing
    fill_tree : bool
        Whether to fill the tree of the destination repository with the
        provided subject and/or visit IDs. Intended to be used when the
        destination repository doesn't contain any of the the input
        filesets/fields (which are stored in external repositories) and
        so the sessions will need to be created in the destination
        repository.


    Class Attrs
    -----------
    add_data_specs : List[FilesetSpec|FieldSpec]
        Adds specs to the '_data_specs' class attribute,
        which is a dictionary that maps the names of filesets that are
        used and generated by the study to (Fileset|Field)Spec objects.
    add_param_specs : List[ParamSpec]
        Adds specs to the '_param_specs' class attribute,
        which is a dictionary that maps the names of parameters that are
        provided to pipelines in the study
    """

    _data_specs = {}
    _param_specs = {}

    implicit_cls_attrs = ['_data_specs', '_param_specs']

    SUBJECT_ID = 'subject_id'
    VISIT_ID = 'visit_id'
    ITERFIELDS = (SUBJECT_ID, VISIT_ID)
    FREQUENCIES = {
        'per_study': (),
        'per_subject': (SUBJECT_ID,),
        'per_visit': (VISIT_ID,),
        'per_session': (SUBJECT_ID, VISIT_ID)}

    def __init__(self, name, repository, processor, inputs,
                 environment=None, parameters=None, subject_ids=None,
                 visit_ids=None, enforce_inputs=True, fill_tree=False,
                 clear_caches=True):
        try:
            # This works for PY3 as the metaclass inserts it itself if
            # it isn't provided
            metaclass = type(self).__dict__['__metaclass__']
            if not issubclass(metaclass, StudyMetaClass):
                raise KeyError
        except KeyError:
            raise ArcanaUsageError(
                "Need to have StudyMetaClass (or a sub-class) as "
                "the metaclass of all classes derived from Study")
        if isinstance(repository, basestring):
            repository = BasicRepo(repository, depth=None)
        if isinstance(processor, basestring):
            processor = SingleProc(processor)
        if environment is None:
            environment = StaticEnv()
        self._name = name
        self._repository = repository
        self._processor = processor.bind(self)
        self._environment = environment
        self._inputs = {}
        self._subject_ids = (tuple(subject_ids)
                             if subject_ids is not None else None)
        self._visit_ids = tuple(visit_ids) if visit_ids is not None else None
        self._fill_tree = fill_tree
        # Initialise caches for data collection and pipeline objects
        if clear_caches:
            self.clear_caches()
        # Set parameters
        if parameters is None:
            parameters = {}
        elif not isinstance(parameters, dict):
            # Convert list of parameters into dictionary
            parameters = {o.name: o for o in parameters}
        self._parameters = {}
        for param_name, param in list(parameters.items()):
            if not isinstance(param, Parameter):
                param = Parameter(param_name, param)
            try:
                param_spec = self._param_specs[param_name]
            except KeyError:
                raise ArcanaNameError(
                    param_name,
                    "Provided parameter '{}' is not present in the "
                    "allowable parameters for {} classes ('{}')"
                    .format(param_name, type(self).__name__,
                            "', '".join(self.param_spec_names())))
            param_spec.check_valid(param, context=' {}(name={})'.format(
                type(self).__name__, name))
            self._parameters[param_name] = param
        # Convert inputs to a dictionary if passed in as a list/tuple
        if not isinstance(inputs, dict):
            inputs = {i.name: i for i in inputs}
        else:
            # Convert string patterns into Input objects
            for inpt_name, inpt in list(inputs.items()):
                if isinstance(inpt, basestring):
                    spec = self.data_spec(inpt_name)
                    if spec.is_fileset:
                        inpt = InputFilesets(inpt_name, pattern=inpt,
                                             is_regex=True)
                    else:
                        inpt = InputFields(inpt_name, pattern=inpt,
                                             dtype=spec.dtype,
                                             is_regex=True)
                    inputs[inpt_name] = inpt
        # Check validity of study inputs
        for inpt_name, inpt in inputs.items():
            try:
                spec = self.data_spec(inpt_name)
            except ArcanaNameError:
                raise ArcanaNameError(
                    inpt.name,
                    "Input name '{}' isn't in data specs of {} ('{}')"
                    .format(
                        inpt.name, self.__class__.__name__,
                        "', '".join(self._data_specs)))
            else:
                if spec.is_fileset:
                    if inpt.is_field:
                        raise ArcanaUsageError(
                            "Passed field ({}) as input to fileset spec"
                            " {}".format(inpt, spec))
                elif not inpt.is_field:
                    raise ArcanaUsageError(
                        "Passed fileset ({}) as input to field spec {}"
                        .format(inpt, spec))
        # "Bind" input selectors to the current study object, and attempt to
        # match with data in the repository
        input_errors = []
        with self.repository:
            if not self.subject_ids:
                raise ArcanaUsageError(
                    "No subject IDs provided and destination repository "
                    "is empty")
            if not self.visit_ids:
                raise ArcanaUsageError(
                    "No visit IDs provided and destination repository "
                    "is empty")
            for inpt_name, inpt in list(inputs.items()):
                try:
                    try:
                        self._inputs[inpt_name] = bound_inpt = inpt.bind(
                            self, spec_name=inpt_name)
                    except ArcanaInputMissingMatchError as e:
                        if not inpt.drop_if_missing:
                            raise e
                    else:
                        spec = self.data_spec(inpt_name)
                        if spec.is_fileset:
                            if spec.derived:
                                try:
                                    spec.format.converter_from(
                                        bound_inpt.format)
                                except ArcanaNoConverterError as e:
                                    e.msg += (
                                        ", which is requried to convert:\n{} "
                                        "to\n{}.".format(e, bound_inpt, spec))
                                    raise e
                            else:
                                if bound_inpt.format not in spec.valid_formats:
                                    raise ArcanaUsageError(
                                        "Cannot pass {} as an input to {} as "
                                        "it is not in one of the valid formats"
                                        " ('{}')".format(
                                            bound_inpt, spec,
                                            "', '".join(
                                                f.name
                                                for f in spec.valid_formats)))
                except ArcanaInputError as e:
                    # Collate errors across all inputs into a single error
                    # message
                    input_errors.append(e)
        if input_errors:
            raise ArcanaInputError('\n'.join(str(e) for e in input_errors))
        # Check remaining specs are optional or have default values
        for spec in self.data_specs():
            if spec.name not in self.input_names:
                if not spec.derived and spec.default is None:
                    # Emit a warning if an acquired fileset has not been
                    # provided for an "acquired fileset"
                    msg = (" acquired fileset '{}' was not given as"
                           " an input of {}.".format(spec.name, self))
                    if spec.optional:
                        logger.info('Optional' + msg)
                    else:
                        if enforce_inputs:
                            raise ArcanaMissingInputError(
                                'Non-optional' + msg + " Pipelines "
                                "depending on this fileset will not "
                                "run")

    def data(self, name, subject_ids=None, visit_ids=None, session_ids=None,
             generate=True, **kwargs):
        """
        Returns the Fileset(s) or Field(s) associated with the provided spec
        name(s), generating derived filesets as required. Multiple names in a
        list can be provided, to allow their workflows to be combined into a
        single workflow.

        Parameters
        ----------
        name : str | List[str]
            The name of the FilesetSpec|FieldSpec to retried the
            filesets for
        subject_id : str | None
            The subject ID of the data to return. If provided (including None
            values) the data will be return as a single item instead of a
            collection
        visit_id : str | None
            The visit ID of the data to return. If provided (including None
            values) the data will be return as a single item instead of a
            c ollection
        subject_ids : list[str]
            The subject IDs to include in the returned collection
        visit_ids : list[str]
            The visit IDs to include in the returned collection
        session_ids : list[str]
            The session IDs (i.e. 2-tuples of the form
            (<subject-id>, <visit-id>) to include in the returned collection
        generate : bool
            Whether to generate data that hasn't been generated yet or not

        Returns
        -------
        data : BaseItem | BaseCollection | list[BaseItem | BaseCollection]
            If 'subject_id' or 'visit_id' is provided then the data returned is
            a single Fileset or Field. Otherwise a collection of Filesets or
            Fields are returned. If muliple spec names are provided then a
            list of items or collections corresponding to each spec name.
        """
        if isinstance(name, basestring):
            single_name = True
            names = [name]
        else:
            names = name
            single_name = False
        single_item = 'subject_id' in kwargs or 'visit_id' in kwargs
        filter_items = (subject_ids, visit_ids, session_ids) != (None, None,
                                                                 None)
        specs = [self.spec(n) for n in names]
        if single_item:
            if filter_items:
                raise ArcanaUsageError(
                    "Cannot provide 'subject_id' and/or 'visit_id' in "
                    "combination with 'subject_ids', 'visit_ids' or "
                    "'session_ids'")
            subject_id = kwargs.pop('subject_id', None)
            visit_id = kwargs.pop('visit_id', None)
            iterators = set(chain(self.FREQUENCIES[s.frequency]
                                  for s in specs))
            if subject_id is not None and visit_id is not None:
                session_ids = [(subject_id, visit_id)]
            elif subject_id is not None:
                if self.VISIT_ID in iterators:
                    raise ArcanaUsageError(
                        "Non-None values for visit IDs need to be "
                        "provided to select a single item for each of '{}'"
                        .format("', '".join(names)))
                subject_ids = [subject_id]
            elif visit_id is not None:
                if self.SUBJECT_ID in iterators:
                    raise ArcanaUsageError(
                        "Non-None values for subject IDs need to be "
                        "provided to select a single item for each of '{}'"
                        .format("', '".join(names)))
                visit_ids = [visit_id]
            elif iterators:
                raise ArcanaUsageError(
                    "Non-None values for subject and/or visit IDs need to be "
                    "provided to select a single item for each of '{}'"
                    .format("', '".join(names)))
        if generate:
            # Work out which pipelines need to be run
            pipeline_getters = defaultdict(set)
            for spec in specs:
                if spec.derived or spec.derivable:  # Filter out Study inputs
                    # Add name of spec to set of required outputs
                    pipeline_getters[(spec.pipeline_getter,
                                      spec.pipeline_args)].add(spec.name)
            # Run required pipelines
            if pipeline_getters:
                kwargs = copy(kwargs)
                kwargs.update({'subject_ids': subject_ids,
                               'visit_ids': visit_ids,
                               'session_ids': session_ids})
                pipelines, required_outputs = zip(*(
                    (self.pipeline(getter, pipeline_args=args), req_outs)
                    for (getter, args), req_outs in pipeline_getters.items()))
                kwargs['required_outputs'] = required_outputs
                self.processor.run(*pipelines, **kwargs)
        # Find and return Item/Collection corresponding to requested spec
        # names
        all_data = []
        for name in names:
            spec = self.bound_spec(name)
            data = spec.collection
            if single_item:
                data = data.item(subject_id=subject_id, visit_id=visit_id)
            elif filter_items and spec.frequency != 'per_study':
                if subject_ids is None:
                    subject_ids = []
                if visit_ids is None:
                    visit_ids = []
                if session_ids is None:
                    session_ids = []
                if spec.frequency == 'per_session':
                    data = [d for d in data
                            if (d.subject_id in subject_ids or
                                d.visit_id in visit_ids or
                                d.session_id in session_ids)]
                elif spec.frequency == 'per_subject':
                    data = [d for d in data
                            if (d.subject_id in subject_ids or
                                d.subject_id in [s[0] for s in session_ids])]
                elif spec.frequency == 'per_visit':
                    data = [d for d in data
                            if (d.visit_id in visit_ids or
                                d.visit_id in [s[1] for s in session_ids])]
                if not data:
                    raise ArcanaUsageError(
                        "No matching data found (subject_ids={}, visit_ids={} "
                        ", session_ids={})"
                        .format(subject_ids, visit_ids, session_ids))
                data = spec.CollectionClass(spec.name, data)
            all_data.append(data)
        if single_name:
            all_data = all_data[0]
        return all_data

    def pipeline(self, getter_name, required_outputs=None, pipeline_args=None):
        """
        Returns a pipeline from a study by getting the method corresponding to
        the given name and checking that the required outputs are generated
        given the parameters of the study

        Parameters
        ----------
        getter_name : str
            Name of the method that constructs the pipeline
        required_outputs : list[str] | None
            The list of outputs that are expected of the pipeline
        pipeline_args : dict[str, *] | None
            Any arguments that should be passed to the method to produce the
            pipeline.
        """
        if pipeline_args is None:
            pipeline_args = ()
        elif isinstance(pipeline_args, dict):
            pipeline_args = tuple(pipeline_args.items())
        try:
            pipeline = self._pipelines_cache[(getter_name, pipeline_args)]
        except KeyError:
            try:
                getter = getattr(self, getter_name)
            except AttributeError:
                raise ArcanaDesignError(
                    "There is no pipeline constructor method named '{}' in "
                    "present in '{}' study".format(getter_name, self))
            self._pipeline_to_generate = getter_name
            try:
                pipeline = getter(**dict(pipeline_args))
            except ArcanaMissingDataException as e:
                e.msg += ("{}, which is required as an input when calling the "
                          "pipeline constructor method '{}' to create a "
                          "pipeline to produce '{}'".format(
                              e, getter_name,
                              "', '".join(required_outputs)))
                raise e
            finally:
                self._pipeline_to_generate = None
            if pipeline is None:
                raise ArcanaDesignError(
                    "'{}' pipeline constructor in {} is missing return "
                    "statement (should return a Pipeline object)".format(
                        getter_name, self))
            elif not isinstance(pipeline, Pipeline):
                raise ArcanaDesignError(
                    "'{}' pipeline constructor in {} doesn't return a Pipeline"
                    " object ({})".format(
                        getter_name, self, pipeline))
            # Check to see if the pipeline is equivalent to previously
            # generated pipelines (if two getter methods return equivalent
            # pipelines) and whether any outputs are to be generated twice
            # by different pipelines within the same workflow
            for prev_pipeline in self._pipelines_cache.values():
                if pipeline == prev_pipeline:
                    pipeline = prev_pipeline
                    break
                elif any(o in prev_pipeline.outputs for o in pipeline.outputs):
                    raise ArcanaDesignError(
                        "'{}' outputs are produced by more than one pipeline "
                        "({} and {})".format(
                            set(pipeline.output_names).intersection(
                                prev_pipeline.output_names),
                            prev_pipeline, pipeline))
            self._pipelines_cache[(getter_name, pipeline_args)] = pipeline
        if required_outputs is not None:
            # Check that the required outputs are created with the given
            # parameters
            missing_outputs = (set(required_outputs) -
                               set(pipeline.output_names))
            if missing_outputs:
                raise ArcanaOutputNotProducedException(
                    "Required output(s) '{}', will not be created by the '{}' "
                    "pipeline constructed by '{}' method in {} given the "
                    "missing study inputs:\n{}\n"
                    "and the provided switches:\n{}"
                    .format(
                        "', '".join(missing_outputs),
                        pipeline.name, getter_name, self,
                        "\n".join(self.missing_inputs),
                        '\n'.join('{}={}'.format(s.name, s.value)
                                  for s in self.switches)))
        return pipeline

    def __repr__(self):
        """String representation of the study"""
        return "{}(name='{}')".format(self.__class__.__name__,
                                      self.name)

    def __reduce__(self):
        """
        Control how study classes are pickled to allow some generated
        classes (those that don't define additional methods) to be
        generated
        """
        cls = type(self)
        module = sys.modules[cls.__module__]
        try:
            # Check whether the study class is generated or not by
            # seeing if it exists in its module
            if cls is not getattr(module, cls.__name__):
                raise AttributeError
        except AttributeError:
            cls_dct = {}
            for name, attr in list(cls.__dict__.items()):
                if isinstance(attr, types.FunctionType):
                    try:
                        if not attr.auto_added:
                            raise ArcanaCantPickleStudyError()
                    except (AttributeError, ArcanaCantPickleStudyError):
                        raise ArcanaCantPickleStudyError(
                            "Cannot pickle auto-generated study class "
                            "as it contains non-auto-added method "
                            "{}:{}".format(name, attr))
                elif name not in self.implicit_cls_attrs:
                    cls_dct[name] = attr
            pkld = (pickle_reconstructor,
                    (cls.__metaclass__, cls.__name__, cls.__bases__,
                     cls_dct), self.__dict__)
        else:
            # Use standard pickling if not a generated class
            pkld = object.__reduce__(self)
        return pkld

    @classmethod
    def static_menu(cls):
        def wrap_desc(desc, nchars, indent='\t\t'):
            lines = []
            while desc:
                if len(desc) > nchars:
                    n = desc[:nchars].rfind(' ')
                else:
                    n = nchars
                lines.append(desc[:n])
                desc = desc[(n + 1):]
            return '\n{}'.format(indent).join(lines)
        cls_name = '.'.join([cls.__module__, cls.__name__])
        menu = ("\n{} Menu \n".format(cls_name) +
                '-' * len(cls_name) + "------")
        menu += '\n\nInputs:'
        for spec in cls.acquired_data_specs():
            if isinstance(spec, BaseFileset):
                menu += '\n\t{} ({}){}:\n\t\t{}'.format(
                    spec.name, ', '.join(f.name for f in spec.valid_formats),
                    (' - optional' if spec.optional else ''),
                    wrap_desc(spec.desc, 80))
            else:
                menu += '\n\t{} ({}{}){}: {}'.format(
                    spec.name, spec.dtype.__name__,
                    (' array' if spec.array else ''),
                    (' - optional' if spec.optional else ''),
                    wrap_desc(spec.desc, 80))
        menu += '\n\nDerivatives:'
        for spec in cls.derived_data_specs():
            if isinstance(spec, BaseFileset):
                menu += '\n\t{} ({}):\n\t\t{}'.format(
                    spec.name, spec.format.name,
                    wrap_desc(spec.desc, 80))
            else:
                menu += '\n\t{} ({}{}): {}'.format(
                    spec.name, spec.dtype.__name__,
                    (' array' if spec.array else ''),
                    wrap_desc(spec.desc, 80))
        menu += '\n\nParameters:'
        for spec in cls.param_specs():
            menu += '\n\t{} ({}){}:\n\t\t{}'.format(
                spec.name, spec.dtype.__name__,
                (' [{}]'.format(", ".join(str(c) for c in spec.choices))
                 if spec.choices else ''),
                wrap_desc(spec.desc, 80))
        return menu

    @property
    def tree(self):
        return self.repository.cached_tree(
            subject_ids=self._subject_ids,
            visit_ids=self._visit_ids,
            fill=self._fill_tree)

    def clear_caches(self):
        """
        Called after a pipeline is run against the study to force an update of
        the derivatives that are now present in the repository if a subsequent
        pipeline is run.
        """
        self.repository.clear_cache()
        self._bound_specs = {}
        self._pipelines_cache = {}

    @property
    def processor(self):
        return self._processor

    @property
    def fill_tree(self):
        return self._fill_tree

    @property
    def environment(self):
        return self._environment

    @property
    def inputs(self):
        return list(self._inputs.values())

    @property
    def input_names(self):
        return list(self._inputs.keys())

    def input(self, name):
        try:
            return self._inputs[name]
        except KeyError:
            raise ArcanaNameError(
                name,
                "{} doesn't have an input named '{}'"
                .format(self, name))

    @property
    def missing_inputs(self):
        return (n for n in self.acquired_data_spec_names()
                if n not in self._inputs)

    @property
    def subject_ids(self):
        if self._subject_ids is None:
            return [s.id for s in self.tree.subjects]
        return self._subject_ids

    @property
    def visit_ids(self):
        if self._visit_ids is None:
            return [v.id for v in self.tree.visits]
        return self._visit_ids

    @property
    def num_subjects(self):
        return len(self.subject_ids)

    @property
    def num_visits(self):
        return len(self.visit_ids)

    @property
    def num_sessions(self):
        if self._visit_ids is None and self._subject_ids is None:
            num_sessions = len(list(self.tree.sessions))
        else:
            num_sessions = self.num_subjects * self.num_visits
        return num_sessions

    @property
    def prefix(self):
        """The study name as a prefix for fileset names"""
        return self.name + '_'

    @property
    def name(self):
        """Accessor for the unique study name"""
        return self._name

    @property
    def repository(self):
        "Accessor for the repository member (e.g. Daris, XNAT, MyTardis)"
        return self._repository

    def new_pipeline(self, *args, **kwargs):
        """
        Creates a Pipeline object, passing the study (self) as the first
        argument
        """
        return Pipeline(self, *args, **kwargs)

    def _get_parameter(self, name):
        try:
            parameter = self._parameters[name]
        except KeyError:
            try:
                parameter = self._param_specs[name]
            except KeyError:
                raise ArcanaNameError(
                    name,
                    "Invalid parameter, '{}', in {} (valid '{}')"
                    .format(
                        name, self._param_error_location,
                        "', '".join(self.param_spec_names())))
        return parameter

    def parameter(self, name):
        """
        Retrieves the value of the parameter and registers the parameter
        as being used by this pipeline for use in provenance capture

        Parameters
        ----------
        name : str
            The name of the parameter to retrieve
        """
        return self._get_parameter(name).value

    def branch(self, name, values=None):  # @UnusedVariable @IgnorePep8
        """
        Checks whether the given switch matches the value provided

        Parameters
        ----------
        name : str
            The name of the parameter to retrieve
        value : str | None
            The value(s) of the switch to match if a non-boolean switch
        """
        if isinstance(values, basestring):
            values = [values]
        spec = self.param_spec(name)
        if not isinstance(spec, SwitchSpec):
            raise ArcanaUsageError(
                "{} is standard parameter not a switch".format(spec))
        switch = self._get_parameter(name)
        if spec.is_boolean:
            if values is not None:
                raise ArcanaDesignError(
                    "Should not provide values ({}) to boolean switch "
                    "'{}' in {}".format(
                        values, name, self._param_error_location))
            in_branch = switch.value
        else:
            if values is None:
                raise ArcanaDesignError(
                    "Value(s) need(s) to be provided non-boolean switch"
                    " '{}' in {}".format(
                        name, self._param_error_location))
            values = set(values)
            # Register parameter as being used by the pipeline
            unrecognised_values = values - set(spec.choices)
            if unrecognised_values:
                raise ArcanaDesignError(
                    "Provided value(s) ('{}') for switch '{}' in {} "
                    "is not a valid option ('{}')".format(
                        "', '".join(unrecognised_values), name,
                        self._param_error_location,
                        "', '".join(spec.choices)))
            in_branch = switch.value in values
            if not in_branch:
                try:
                    in_branch = spec.fallbacks[switch.value] in values
                except KeyError:
                    pass
        return in_branch

    def unhandled_branch(self, name):
        """
        Convenient method for raising exception if a pipeline doesn't
        handle a particular switch value

        Parameters
        ----------
        name : str
            Name of the switch
        value : str
            Value of the switch which hasn't been handled
        """
        raise ArcanaDesignError(
            "'{}' value of '{}' switch in {} is not handled"
            .format(self.parameter(name), name,
                    self._param_error_location))

    @property
    def _param_error_location(self):
        return ("generation of '{}' pipeline of {}"
                .format(self._pipeline_to_generate, self))

    @property
    def parameters(self):
        for param_name in self._param_specs:
            yield self._get_parameter(param_name)

    @property
    def switches(self):
        for name in self._param_specs:
            if isinstance(self.spec(name), SwitchSpec):
                yield self._get_parameter(name)

    def save_workflow_graph_for(self, spec_name, fname, full=False,
                                style='flat', **kwargs):
        """
        Saves a graph of the workflow to generate the requested spec_name

        Parameters
        ----------
        spec_name : str
            Name of the spec to generate the graph for
        fname : str
            The filename for the saved graph
        style : str
            The style of the graph, can be one of can be one of
            'orig', 'flat', 'exec', 'hierarchical'
        """
        pipeline = self.spec(spec_name).pipeline
        if full:
            workflow = pe.Workflow(name='{}_gen'.format(spec_name),
                                   base_dir=self.processor.work_dir)
            self.processor._connect_pipeline(
                pipeline, workflow, **kwargs)
        else:
            workflow = pipeline._workflow
        fname = op.expanduser(fname)
        if not fname.endswith('.png'):
            fname += '.png'
        dotfilename = fname[:-4] + '.dot'
        workflow.write_graph(graph2use=style,
                             dotfilename=dotfilename)

    def spec(self, name):
        """
        Returns either the input corresponding to a fileset or field
        field spec or a spec or parameter that has either
        been passed to the study as an input or can be derived.

        Parameters
        ----------
        name : Str | BaseData | Parameter
            A parameter, fileset or field or name of one
        """
        # If the provided "name" is actually a data item or parameter then
        # replace it with its name.
        if isinstance(name, (BaseData, Parameter)):
            name = name.name
        # If name is a parameter than return the parameter spec
        if name in self._param_specs:
            return self._param_specs[name]
        else:
            return self.bound_spec(name)

    def bound_spec(self, name):
        """
        Returns an input selector or derived spec bound to the study, i.e.
        where the repository tree is checked for existing outputs

        Parameters
        ----------
        name : Str
            A name of a fileset or field
        """
        # If the provided "name" is actually a data item or parameter then
        # replace it with its name.
        if isinstance(name, BaseData):
            name = name.name
        # Get the spec from the class
        spec = self.data_spec(name)
        try:
            bound = self._inputs[name]
        except KeyError:
            if not spec.derived and spec.default is None:
                raise ArcanaMissingDataException(
                    "Input (i.e. non-generated) data '{}' "
                    "was not supplied when the study '{}' was "
                    "initiated".format(name, self.name))
            else:
                try:
                    bound = self._bound_specs[name]
                except KeyError:
                    bound = self._bound_specs[name] = spec.bind(self)
        return bound

    @classmethod
    def data_spec(cls, name):
        """
        Return the fileset_spec, i.e. the template of the fileset expected to
        be supplied or generated corresponding to the fileset_spec name.

        Parameters
        ----------
        name : Str
            Name of the fileset_spec to return
        """
        # If the provided "name" is actually a data item or parameter then
        # replace it with its name.
        if isinstance(name, BaseData):
            name = name.name
        try:
            return cls._data_specs[name]
        except KeyError:
            raise ArcanaNameError(
                name,
                "No fileset spec named '{}' in {}, available:\n{}"
                .format(name, cls.__name__,
                        "\n".join(list(cls._data_specs.keys()))))

    @classmethod
    def param_spec(cls, name):
        try:
            return cls._param_specs[name]
        except KeyError:
            raise ArcanaNameError(
                name,
                "No parameter spec named '{}' in {}, available:\n{}"
                .format(name, cls.__name__,
                        "\n".join(list(cls._param_specs.keys()))))

    @classmethod
    def data_specs(cls):
        """Lists all data_specs defined in the study class"""
        return iter(cls._data_specs.values())

    @classmethod
    def param_specs(cls):
        return iter(cls._param_specs.values())

    @classmethod
    def data_spec_names(cls):
        """Lists the names of all data_specs defined in the study"""
        return iter(cls._data_specs.keys())

    @classmethod
    def param_spec_names(cls):
        """Lists the names of all param_specs defined in the study"""
        return iter(cls._param_specs.keys())

    @classmethod
    def spec_names(cls):
        return chain(cls.data_spec_names(),
                     cls.param_spec_names())

    @classmethod
    def acquired_data_specs(cls):
        """
        Lists all data_specs defined in the study class that are
        provided as inputs to the study
        """
        return (c for c in cls.data_specs() if not c.derived)

    @classmethod
    def derived_data_specs(cls):
        """
        Lists all data_specs defined in the study class that are typically
        generated from other data_specs (but can be overridden by input
        filesets)
        """
        return (c for c in cls.data_specs() if c.derived)

    @classmethod
    def derived_data_spec_names(cls):
        """Lists the names of generated data_specs defined in the study"""
        return (c.name for c in cls.derived_data_specs())

    @classmethod
    def acquired_data_spec_names(cls):
        "Lists the names of acquired data_specs defined in the study"
        return (c.name for c in cls.acquired_data_specs())

    def cache_inputs(self):
        """
        Runs the Study's repository source node for each of the inputs
        of the study, thereby caching any data required from remote
        repositorys. Useful when launching many parallel jobs that will
        all try to concurrently access the remote repository, and probably
        lead to timeout errors.
        """
        workflow = pe.Workflow(name='cache_download',
                               base_dir=self.processor.work_dir)
        subjects = pe.Node(IdentityInterface(['subject_id']), name='subjects',
                           environment=self.environment)
        sessions = pe.Node(IdentityInterface(['subject_id', 'visit_id']),
                           name='sessions', environment=self.environment)
        subjects.iterables = ('subject_id', tuple(self.subject_ids))
        sessions.iterables = ('visit_id', tuple(self.visit_ids))
        source = pe.Node(RepositorySource(
            self.bound_spec(i).collection for i in self.inputs), name='source')
        workflow.connect(subjects, 'subject_id', sessions, 'subject_id')
        workflow.connect(sessions, 'subject_id', source, 'subject_id')
        workflow.connect(sessions, 'visit_id', source, 'visit_id')
        workflow.run()

    @classmethod
    def print_specs(cls):
        print('Available data:')
        for spec in cls.data_specs():
            print(spec)
        print('\nAvailable parameters:')
        for spec in cls.param_specs():
            print(spec)

    def provided(self, spec_name, default_okay=True):
        """
        Checks to see whether the corresponding data spec was provided an
        explicit input, as opposed to derivatives or missing optional inputs

        Parameters
        ----------
        spec_name : str
            Name of a data spec
        default_okay : bool
            Whether or not the default value is okay, or whether it needs to
            be explicitly provided as an input. For derivatives the default is
            the derived value
        """
        try:
            spec = self.bound_spec(spec_name)
        except ArcanaMissingDataException:
            return False
        if isinstance(spec, BaseInput):
            return True
        elif default_okay:
            if isinstance(spec, BaseInputSpec):
                return spec.default is not None
            else:
                return True
        else:
            return False

    @classmethod
    def freq_from_iterators(cls, iterators):
        """
        Returns the frequency corresponding to the given iterators
        """
        return {
            set(it): f for f, it in cls.FREQUENCIES.items()}[set(iterators)]

    @property
    def prov(self):
        """
        Extracts provenance information from the study for storage alongside
        generated derivatives. Typically for reference purposes only as only
        the pipeline workflow, inputs and outputs are checked by default when
        determining which sessions require reprocessing.

        Returns
        -------
        prov : dict[str, *]
            A dictionary containing the provenance information to record
            for the study
        """
        # Get list of repositories where inputs to the study are stored
        input_repos = list(set((i.repository for i in self.inputs)))
        inputs = {}
        for input in self.inputs:  # @ReservedAssignment
            inputs[input.name] = {
                'repository_index': input_repos.index(input.repository)}
            if input.frequency == 'per_study':
                inputs[input.name]['names'] = next(input.collection).name
            elif input.frequency == 'per_subject':
                inputs[input.name]['names'] = {i.subject_id: i.name
                                               for i in input.collection}
            elif input.frequency == 'per_visit':
                inputs[input.name]['names'] = {i.visit_id: i.name
                                               for i in input.collection}
            elif input.frequency == 'per_session':
                names = defaultdict(dict)
                for item in input.collection:
                    names[item.subject_id][item.visit_id] = item.name
                # Convert from defaultdict to dict
                inputs[input.name]['names'] = dict(names.items())
        return {
            'name': self.name,
            'type': get_class_info(type(self)),
            'parameters': {p.name: p.value for p in self.parameters},
            'inputs': inputs,
            'environment': self.environment.prov,
            'repositories': [r.prov for r in input_repos],
            'processor': self.processor.prov,
            'subject_ids': tuple(self.subject_ids),
            'visit_ids': tuple(self.visit_ids)}


class StudyMetaClass(type):
    """
    Metaclass for all study classes that collates data specs from
    bases and checks pipeline method names.

    Combines specifications in add_(data|parameter)_specs from
    the class to be created with its base classes, overriding matching
    specs in the order of the bases.
    """

    def __new__(metacls, name, bases, dct):  # @NoSelf @UnusedVariable
        if not any(issubclass(b, Study) for b in bases):
            raise ArcanaUsageError(
                "StudyMetaClass can only be used for classes that "
                "have Study as a base class")
        try:
            add_data_specs = dct['add_data_specs']
        except KeyError:
            add_data_specs = []
        try:
            add_param_specs = dct['add_param_specs']
        except KeyError:
            add_param_specs = []
        combined_attrs = set()
        combined_data_specs = {}
        combined_param_specs = {}
        for base in reversed(bases):
            # Get the combined class dictionary including base dicts
            # excluding auto-added properties for data and parameter specs
            combined_attrs.update(
                a for a in dir(base) if (not issubclass(base, Study) or
                                         a not in base.spec_names()))
            # TODO: need to check that fields are not overridden by filesets
            #       and vice-versa
            try:
                combined_data_specs.update(
                    (d.name, d) for d in base.data_specs())
            except AttributeError:
                pass  # Not a Study class
            try:
                combined_param_specs.update(
                    (p.name, p) for p in base.param_specs())
            except AttributeError:
                pass  # Not a Study class
        combined_attrs.update(list(dct.keys()))
        combined_data_specs.update((d.name, d) for d in add_data_specs)
        combined_param_specs.update(
            (p.name, p) for p in add_param_specs)
        # Check that the pipeline names in data specs correspond to a
        # pipeline method in the class and that if a pipeline is called with
        # arguments (for parameterizing a range of metrics for example) then
        # the same argument names are consistent across every call.
        pipeline_arg_names = {}
        for spec in add_data_specs:
            if spec.derived:
                if spec.pipeline_getter in ('pipeline', 'new_pipeline'):
                    raise ArcanaDesignError(
                        "Cannot use the names 'pipeline' or  'new_pipeline' "
                        "('{}') for the name of a pipeline constructor in "
                        "class {} as it clashes with base method to create "
                        "pipelines"
                        .format(spec.pipeline_getter, name))
                if spec.pipeline_getter not in combined_attrs:
                    raise ArcanaDesignError(
                        "Pipeline to generate '{}', '{}', is not present"
                        " in '{}' class".format(
                            spec.name, spec.pipeline_getter, name))
                try:
                    if pipeline_arg_names[
                            spec.pipeline_getter] != spec.pipeline_arg_names:
                        raise ArcanaDesignError(
                            "Inconsistent pipeline argument names used for "
                            "'{}' pipeline getter {} and {}".format(
                                spec.pipeline_getter,
                                pipeline_arg_names[spec.pipeline_getter],
                                spec.pipeline_arg_names))
                except KeyError:
                    pipeline_arg_names[
                        spec.pipeline_getter] = spec.pipeline_arg_names
        # Check for name clashes between data and parameter specs
        spec_name_clashes = (set(combined_data_specs) &
                             set(combined_param_specs))
        if spec_name_clashes:
            raise ArcanaDesignError(
                "'{}' name both data and parameter specs in '{}' class"
                .format("', '".join(spec_name_clashes), name))
        reserved_clashes = [n for n in combined_data_specs
                            if n in Study.ITERFIELDS]
        if reserved_clashes:
            raise ArcanaDesignError(
                "'{}' data spec names clash with reserved names"
                .format("', '".join(reserved_clashes), name))
        dct['_data_specs'] = combined_data_specs
        dct['_param_specs'] = combined_param_specs
        if '__metaclass__' not in dct:
            dct['__metaclass__'] = metacls
        # Append description of Study parameters to class
        try:
            docstring = dct['__doc__']
        except KeyError:
            docstring = '{} Study class'.format(name)
        if 'Parameters' not in docstring:
            docstring += Study.__doc__
        dct['__doc__'] = docstring
        return type(name, bases, dct)


def pickle_reconstructor(metacls, name, bases, cls_dict):
    obj = DummyObject()
    obj.__class__ = metacls(name, bases, cls_dict)
    return obj


class DummyObject(object):
    pass
